Module LCD

Include "utils.bas"
Include "system.bas"

// by default, the module calls SetAllDigital in Initialize()
// set option false to exclude the call
#if IsOption(LCD_SETALLDIGITAL) And Not (LCD_SETALLDIGITAL in (true, false))
   #error LCD_SETALLDIGITAL, "Invalid option"
#endif
#option LCD_SETALLDIGITAL = true

' option section - if no values are given in the main program, this section
' will initialise them - all values are verified and if any errors are found,
' a custom message will be displayed in the main user program...

' LCD interface...
#option LCD_INTERFACE = 4           ' can be 4 or 8 bit interface
#if IsOption(LCD_INTERFACE) And Not (LCD_INTERFACE in (4, 8))
   #error LCD_INTERFACE, "Invalid option. LCD_INTERFACE must be either 4 or 8."
#endif

' command timeout...
#option LCD_COMMAND_US = 2000       ' command delay in micro seconds
#if IsOption(LCD_COMMAND_US) And Not (LCD_COMMAND_US in (1 to 65535))
   #error LCD_COMMAND_US, "Invalid option. Command delay must be between 1 and 65535 us."
#endif

' data timeout...
#option LCD_DATA_US = 50            ' data delays in micro seconds
#if IsOption(LCD_DATA_US) And Not (LCD_DATA_US in (1 to 255))
   #error LCD_DATA_US, "Invalid option. _data delay must be between 1 and 255 us."
#endif

' module initialsation wait...
#option LCD_INIT_DELAY = 100        ' module startup delay in milliseconds
#if IsOption(LCD_INIT_DELAY) And Not (LCD_INIT_DELAY in (0 to 1000))
   #error LCD_INIT_DELAY, "Invalid option. LCD initialize delay must be between 0 and 1000 (ms)."
#endif 

' RS pin...
#option LCD_RS = PORTB.3
#if IsOption(LCD_RS) And Not IsValidPortPin(LCD_RS) 
   #error LCD_RS, "Invalid option. RS must be a valid pin."
#endif

' EN pin...
#option LCD_EN = PORTB.2
#if IsOption(LCD_EN) And Not IsValidPortPin(LCD_EN) 
   #error LCD_EN, "Invalid option. EN must be a valid pin."
#endif

' standard DDRAM line offsets - should really only be set
' by other modules which build on this LCD library
#option LCD_LINE_2 = $40
#option LCD_LINE_3 = $14
#option LCD_LINE_4 = $54
Private Const LCDLine_2 = LCD_LINE_2
Private Const LCDLine_3 = LCD_LINE_3
Private Const LCDLine_4 = LCD_LINE_4

' the default data pins for 8 bit mode...
Public Dim D0 As PORTC.7
Public Dim D1 As PORTC.6
Public Dim D2 As PORTB.7
Public Dim D3 As PORTB.6
Public Dim D4 As PORTC.0
Public Dim D5 As PORTC.1
Public Dim D6 As PORTC.2
Public Dim D7 As PORTB.5
Public Dim D8 As PORTB.4

#if LCD_INTERFACE = 8
#option LCD_D0 = PORTC.7
#option LCD_D1 = PORTC.6
#option LCD_D2 = PORTC.5
#option LCD_D3 = PORTC.4
#option LCD_D4 = PORTC.3
#option LCD_D5 = PORTC.2
#option LCD_D6 = PORTC.1
#option LCD_D7 = PORTC.0

' default pins for 4 bit mode...
#else
#option LCD_D0 = PORTC.3
#option LCD_D1 = PORTC.2
#option LCD_D2 = PORTC.1
#option LCD_D3 = PORTC.0
#endif

' validate data pins...
#if IsOption(LCD_D0) And Not IsValidPortPin(LCD_D0) 
   #error LCD_D0, "Invalid option. LCD_D0 must be a valid pin."
#endif
#if IsOption(LCD_D1) And Not IsValidPortPin(LCD_D1) 
   #error LCD_D1, "Invalid option. LCD_D1 must be a valid pin."
#endif
#if IsOption(LCD_D2) And Not IsValidPortPin(LCD_D2) 
   #error LCD_D2, "Invalid option. LCD_D2 must be a valid pin."
#endif
#if IsOption(LCD_D3) And Not IsValidPortPin(LCD_D3) 
   #error LCD_D3, "Invalid option. LCD_D3 must be a valid pin."
#endif
#if LCD_INTERFACE = 8
#if IsOption(LCD_D4) And Not IsValidPortPin(LCD_D4) 
   #error LCD_D4, "Invalid option. LCD_D4 must be a valid pin."
#endif
#if IsOption(LCD_D5) And Not IsValidPortPin(LCD_D5) 
   #error LCD_D5, "Invalid option. LCD_D5 must be a valid pin."
#endif
#if IsOption(LCD_D6) And Not IsValidPortPin(LCD_D6) 
   #error LCD_D6, "Invalid option. LCD_D6 must be a valid pin."
#endif
#if IsOption(LCD_D7) And Not IsValidPortPin(LCD_D7) 
   #error LCD_D7, "Invalid option. LCD_D7 must be a valid pin."
#endif
#endif ' LCD_INTERFACE = 8

' by default, the module assumes RW is not used and tied to ground - if the
' user specifies a RW pin, then data read code is linked in...
#if IsOption(LCD_RW)
   #define _RW_USED
   #if Not IsValidPortPin(LCD_RW) 
      #error LCD_RW, "Invalid option. RW must be a valid pin."
   #endif   
#endif

' public commands...
Public Const cmdCGRAM     = %01000000
Public Const cmdDDRAM     = %10000000
Public Const cmdClear     = %00000001
Public Const cmdHome      = %00000010
Public Const cmdCursorOff = %00001100
Public Const cmdCursorOn  = %00001110  
Public Const cmdBlinkOn   = %00001101
Public Const cmdBlinkOff  = %00001100
Public Const cmdMoveLeft  = %00010000
Public Const cmdMoveRight = %00010100     

' bring delay options into the module...
Private Const _dataUS As Byte      = LCD_DATA_US
Private Const _commandUS As Word = LCD_COMMAND_US
Private Const _LcdDelay As Word  = LCD_INIT_DELAY  

' bring pin options into the module...
Private Dim _data_0 As LCD_D0.LCD_D0@
Private Dim _data_1 As LCD_D1.LCD_D1@
Private Dim _data_2 As LCD_D2.LCD_D2@
Private Dim _data_3 As LCD_D3.LCD_D3@
#if LCD_INTERFACE = 8
Private Dim _data_4 As LCD_D4.LCD_D4@
Private Dim _data_5 As LCD_D5.LCD_D5@
Private Dim _data_6 As LCD_D6.LCD_D6@
Private Dim _data_7 As LCD_D7.LCD_D7@
#endif
Private Dim _rs As LCD_RS.LCD_RS@       ' register select
Private Dim _en As LCD_EN.LCD_EN@       ' enable read
Private Dim _posX, _posY As Byte        ' local x, y coordinates  

' RW specified...
#ifdef _RW_USED
Private Dim _rw As LCD_RW.LCD_RW@
#endif

'****************************************************************************
'* Name    : StrobeEN                                                       *
'* Purpose : Strobe read enable - assumes _en is in zero state. Rather than *
'*         : halt the processor after _en = 0, the function returns         *
'*         : immediately                                                    *
'****************************************************************************
Private Inline Sub StrobeEN()
  _en = 1
  DelayUS(1)
  _en = 0
End Sub

'****************************************************************************
'* Name    : MakeOutput                                                     *
'* Purpose : Make data pins output                                          *
'****************************************************************************
Private Inline Sub MakeOutput()
   Output(_data_0)
   Output(_data_1)
   Output(_data_2)
   Output(_data_3)
   
   #if LCD_INTERFACE = 8
   Output(_data_4)
   Output(_data_5)
   Output(_data_6)
   Output(_data_7)   
   #endif
End Sub

'****************************************************************************
'* Name    : MakeInput                                                      *
'* Purpose : Make data pins input                                           *
'****************************************************************************
#ifdef _RW_USED
Private Inline Sub MakeInput()
   Input(_data_0)
   Input(_data_1)
   Input(_data_2)
   Input(_data_3)
   
   #if LCD_INTERFACE = 8
   Input(_data_4)
   Input(_data_5)
   Input(_data_6)
   Input(_data_7)   
   #endif
End Sub

'****************************************************************************
'* Name    : GetData                                                        *
'* Purpose : Read _data port value. For 4 bit interface, value is returned  *
'*         : in LSNibble                                                    *
'****************************************************************************
Private Function GetData() As Byte
   _en = 1 
   DelayUS(1)

   GetData = 0
   GetData.0 = _data_0
   GetData.1 = _data_1
   GetData.2 = _data_2
   GetData.3 = _data_3
   
   #if LCD_INTERFACE = 8
   GetData.4 = _data_4
   GetData.5 = _data_5
   GetData.6 = _data_6
   GetData.7 = _data_7
   #endif
   _en = 0 
End Function

'****************************************************************************
'* Name    : ReadByte                                                       *
'* Purpose : Read a data byte from the device                               *
'****************************************************************************
Private Function ReadByte() As Byte
   MakeInput()
   _rs = 0
   _rw = 1
   #if LCD_INTERFACE = 8
   ReadByte = GetData
   #else   
   ReadByte = GetData << 4                    ' high nibble
   ReadByte = ReadByte Or (GetData And $0F)  ' low nibble
   #endif
   _rw = 0
   MakeOutput()
End Function

'****************************************************************************
'* Name    : WaitFor                                                        *
'* Purpose : Wait for device busy flag to clear                             *
'****************************************************************************
Private Sub WaitFor()
   Repeat
   Until (ReadByte() And $80) = 0
End Sub  
#endif ' _RW_USED

'****************************************************************************
'* Name    : SetData                                                        *
'* Purpose : Sets DATA port value                                           *
'****************************************************************************
Private Sub SetData(data As Byte)
   _data_0 = data.0
   _data_1 = data.1
   _data_2 = data.2
   _data_3 = data.3
   
   #if LCD_INTERFACE = 8
   _data_4 = data.4
   _data_5 = data.5
   _data_6 = data.6
   _data_7 = data.7
   #endif  
   StrobeEN()
   ClrWDT()
End Sub

'****************************************************************************
'* Name    : SetDDRAM                                                       *
'* Purpose : Set DDRAM address                                              *
'****************************************************************************
Private Sub SetDDRAM(address As Byte)
   #ifdef _RW_USED
   WaitFor()
   #endif
   
   _rs = 0
   #if LCD_INTERFACE = 8
   SetData(address Or %10000000)
   #else
   SetData((address Or %10000000) >> 4)
   SetData(address)
   #endif  
   
   #ifndef _RW_USED
   DelayUS(_dataUS) 
   #endif
   ClrWDT()
End Sub

'****************************************************************************
'* Name    : WriteCommandAsByte                                             *
'* Purpose : Write byte command to LCD                                      *
'****************************************************************************
Private Sub WriteCommandAsByte(cmd As Byte)
   #ifdef _RW_USED
   WaitFor()
   #endif
   
   _rs = 0
   #if LCD_INTERFACE = 8
   SetData(cmd)
   #else
   SetData(cmd >> 4)
   SetData(cmd)
   #endif   
   
   #ifndef _RW_USED
   DelayUS(_commandUS)
   #endif
   ClrWDT()
End Sub

'****************************************************************************
'* Name    : WriteCommand                                                   *
'* Purpose : Write command to LCD                                           *
'****************************************************************************
Public Inline Sub WriteCommand(cmd As Byte)
   WriteCommandAsByte(cmd)
End Sub

'****************************************************************************
'* Name    : SetCursor                                                      *
'* Purpose : Set the cursor to line and column                              *
'****************************************************************************
Public Sub SetCursor(line As Byte, column As Byte)
   Dec(column)
   Select line
      Case 1 : SetDDRAM(column)
      Case 2 : SetDDRAM(LCDLine_2 + column)
      Case 3 : SetDDRAM(LCDLine_3 + column)
      Case 4 : SetDDRAM(LCDLine_4 + column)
   End Select 
End Sub

'****************************************************************************
'* Name    : WriteItem                                                      *
'* Purpose : Write a single byte to the LCD                                 *
'****************************************************************************
Private Sub WriteItem(data As Byte)
   #ifdef _RW_USED
   WaitFor()
   #endif
   
   _rs = 1
   #if LCD_INTERFACE = 8
   SetData(data)
   #else
   SetData(data >> 4)
   SetData(data)
   #endif   
   
   #ifndef _RW_USED
   DelayUS(_dataUS) 
   #endif
   ClrWDT()
End Sub

'****************************************************************************
'* Name    : WriteItem                                                      *
'* Purpose : Write a string to the LCD                                      *
'****************************************************************************
Private Sub WriteItem(pText As String)
   Dim TextPtr As POSTINC0
   Dim Text As INDF0
   FSR0 = AddressOf(pText)
   While Text <> 0
      WriteItem(TextPtr)
   Wend
End Sub

'****************************************************************************
'* Name    : WriteItem                                                      *
'* Purpose : Initialise the CGRAM with constant data array. Eight           *
'*         : programmable characters are available (0..7).                  *
'****************************************************************************
Private Sub WriteItem(ByRefConst pBitmap() As Byte)
   Dim Index As Byte
   WriteCommand(cmdCGRAM)
   For Index = 0 To Bound(pBitmap)
      WriteItem(pBitmap(Index))
   Next
   WriteCommand(cmdDDRAM)
End Sub

'****************************************************************************
'* Name    : Write                                                          *
'* Purpose : Calls one or more WriteItem() subroutines                      *
'****************************************************************************
Public Compound Sub Write(WriteItem)

'****************************************************************************
'* Name    : MoveTo                                                         *
'* Purpose : Moves cursor to module private x, y location                   *
'****************************************************************************
Private Inline Sub MoveTo()
   SetCursor(_posY, _posX)
End Sub 

'****************************************************************************
'* Name    : SetLocationX                                                   *
'* Purpose : Set cursor x                                                   *
'****************************************************************************
Private Inline Sub SetLocationX(x As _posX)
End Sub

'****************************************************************************
'* Name    : SetLocationY                                                   *
'* Purpose : Set cursor y                                                   *
'****************************************************************************
Private Inline Sub SetLocationY(y As _posY)
End Sub

'****************************************************************************
'* Name    : WriteAt                                                        *
'* Purpose : Calls one or more WriteItem() subroutines at location x, y     *
'****************************************************************************
Public Compound Sub WriteAt(SetLocationY, SetLocationX, MoveTo, WriteItem)  

'****************************************************************************
'* Name    : Cls                                                            *
'* Purpose : Clear the LCD screen                                           *
'****************************************************************************
Public Sub Cls()
   WriteCommand(cmdClear)
   DelayMS(30)
End Sub

'****************************************************************************
'* Name    : DoInitSequence                                                 *
'* Purpose : Performs a initialization sequence                             *
'****************************************************************************
Private Sub DoInitSequence(value As Byte)
   SetData(value)
   DelayMS(5)
   SetData(value)
   DelayMS(5)
   SetData(value)
   DelayUS(150)
End Sub 

'****************************************************************************
'* Name    : Initialize                                                     *
'* Purpose : Initialize the LCD                                             *
'****************************************************************************
Private Sub Initialize()

   DelayMS(_LcdDelay)
   
   _posX = 1
   _posY = 1
   
   #if (LCD_SETALLDIGITAL)
   SetAllDigital()
   #endif
   
   ' read write...
   #ifdef _RW_USED
   Low(_rw)
   #endif

   ' set _rs and EN to low output...
   Low(_rs)
   Low(_en)
   DelayMS(15)  
   
   ' initialise for 8 bit interface...
   #if LCD_INTERFACE = 8
   MakeOutput
   DoInitSequence($30)
   WriteCommandAsByte(%00111000)    ' DL = 1, N = 1, F = 0
   
   ' initialise for 4 bit interface...
   #else
   MakeOutput
   DoInitSequence($03)
   SetData($02) 
   DelayUS(160) 
   WriteCommandAsByte(%00101000)    ' DL = 0, N = 1, F = 0
   #endif  
   
   WriteCommandAsByte(%00001000)    ' D = 0, C = 0, B = 0
   WriteCommandAsByte(%00000001)    ' Clear
   WriteCommandAsByte(%00000110)    ' ID = 1, S = 0
   WriteCommandAsByte(%00001100)    ' D = 1, C = 0, B = 0 
End Sub

// initialise the module...
Initialize()

End Module

