Module Dogm

#option LCD_DOGM = DOGM_162
#if IsOption(LCD_DOGM) And Not (LCD_DOGM in (DOGM_161, DOGM_162, DOGM_163))
   #error LCD_DOGM, "Invalid option. DOGM type must DOGM_161, DOGM_162 or DOGM_163."
#endif 

#if LCD_DOGM = DOGM_163
#option LCD_LINE_2 = $10
#option LCD_LINE_3 = $20
#endif

#option LCD_BACKLIGHT = PORTB.6
#if IsOption(LCD_BACKLIGHT) And Not IsValidPortPin(LCD_BACKLIGHT) 
   #error LCD_BACKLIGHT, "Invalid option.  LCD_BACKLIGHT must be a valid port pin."
#endif

' LCD interface
Include "Dogm.bas"

' import LCD constants...
Public Const cmdCGRAM     = LCD.cmdCGRAM
Public Const cmdDDRAM     = LCD.cmdDDRAM
Public Const cmdClear     = LCD.cmdClear
Public Const cmdHome      = LCD.cmdHome
Public Const cmdCursorOff = LCD.cmdCursorOff
Public Const cmdCursorOn  = LCD.cmdCursorOn 
Public Const cmdBlinkOn   = LCD.cmdBlinkOn
Public Const cmdBlinkOff  = LCD.cmdBlinkOff
Public Const cmdMoveLeft  = LCD.cmdMoveLeft
Public Const cmdMoveRight = LCD.cmdMoveRight

' import LCD subs...
Public Dim Cls As LCD.Cls
Public Dim SetCursor As LCD.SetCursor
Public Dim Write As LCD.Write
Public Dim WriteAt As LCD.WriteAt
Public Dim WriteCommand As LCD.WriteCommand 

Public Const minBacklight = 0
Public Const maxContrast = 32
Public Const minContrast = 0

' private variables
Dim _backlightPin As LCD_BACKLIGHT.LCD_BACKLIGHT@
Dim _backlight As Word
Dim _contrast As Byte

'****************************************************************************
'* Name    : SetBacklight                                                   *
'* Purpose : Set the LCD Plus backlight                                     *
'****************************************************************************
Public Sub SetBacklight(value As Word)
   If value > 0 Then
      High(_backlightPin)
   Else
      Low(_backlightPin)
   End If
End Sub 

'****************************************************************************
'* Name    : GetBacklight                                                   *
'* Purpose : Get the LCD Plus backlight value                               *
'****************************************************************************
Public Function GetBacklight() As Word
   result = _backlight
End Function 

'****************************************************************************
'* Name    : SetContrast                                                    *
'* Purpose : Set the LCD Plus contrast                                      *
'****************************************************************************
Public Sub SetContrast(value As Byte)
   If value <= maxContrast Then 
      _contrast = value
   End If
   LCD.WriteCommand($70 + _contrast)  
End Sub 

'****************************************************************************
'* Name    : GetContrast                                                    *
'* Purpose : Get the LCD Plus contrast value                                *
'****************************************************************************
Public Function GetContrast() As Byte
   result = _contrast
End Function 

'****************************************************************************
'* Name    : Initialize                                                     *
'* Purpose : Initialize the LCD                                             *
'****************************************************************************
Private Sub Initialize()
   _backlight = 0
   _contrast = 7

   ' DOGM display - 5.0v startup sequence...   
   ' http://www.lcd-module.com/eng/pdf/doma/dog-me.pdf  
   #if LCD_DOGM = DOGM_162
   LCD.WriteCommand($29)      ' 4 bit interface     
   LCD.WriteCommand($1C)      ' bias set
   LCD.WriteCommand($52)      ' power control
   LCD.WriteCommand($69)      ' follower control 
   #else
   LCD.WriteCommand($29)      ' 4 bit interface     
   LCD.WriteCommand($1D)      ' bias set
   LCD.WriteCommand($50)      ' power control
   LCD.WriteCommand($6C)      ' follower control   
   #endif 
   
   ' set contrast to default value...  
   SetContrast(_contrast)   
End Sub

// initialise the module...
Initialize()

End Module

